#include <GL/glew.h>
#include <QImage.h>
#include <QGL.h>
#include <vcg/Point3.h>
#include "MMGrid.h"

using namespace std;
using namespace vcg;

void MMGrid::drawGL()
{
	glPushMatrix();
	float s=2.0f/max(sx,sy);
	glTranslatef(-1.0f,-1.0f,0);
	glScalef(s,s,s);
	if(ti) 
	{
		glEnable(GL_TEXTURE_2D);
	  glBindTexture(GL_TEXTURE_2D,ti);
	}
	vector<MMCell>::iterator ci;
	glGetDoublev(GL_MODELVIEW_MATRIX,mm); 
	glGetDoublev(GL_PROJECTION_MATRIX,mp); 
	glGetIntegerv(GL_VIEWPORT,vp);
	for(ci=g.begin();ci!=g.end();++ci)
    (*ci).drawGL();
	glPopMatrix();
}


bool MMGrid::pickGL(Point2i in, Point2i &pos)
{
	float z;
	glReadPixels(in[0],vp[3]-in[1],1,1,GL_DEPTH_COMPONENT ,GL_FLOAT,&z);
	return pickGL(Point3d(in[0],vp[3]-in[1],z),pos);
}

bool MMGrid::pickGL(Point3d in, Point2i &pos)
{
	Point3d o;
	gluUnProject(in[0],in[1],in[2],mm,mp,vp,&o[0],&o[1],&o[2]);
  //if(fabs(o[2] ) < .1) 
	{
		qDebug("picked %f %f %f\n",o[0],o[1],o[2]);
		if(o[0]<0 || o[0]>sx ) return false;
		if(o[1]<0 || o[1]>sy ) return false;
		pos[0]=o[0];
		pos[1]=o[1]; 
		
		return true;
	}
	return false;
}

void MMGrid::resize(const int _sx, const int _sy) 
{
	sx=_sx; sy=_sy;
	g.resize(sx*sy);
	int i,j;
	for(i=0;i<sx;++i)
		for(j=0;j<sy;++j)
		{
			G(i,j).V[0]=Point3f(i+0,j+0,0);
			G(i,j).V[1]=Point3f(i+1,j+0,0);
			G(i,j).V[2]=Point3f(i+1,j+1,0);
			G(i,j).V[3]=Point3f(i+0,j+1,0);
			G(i,j).ComputeNormal();
			G(i,j).selected=false;

		}
}
// in una cella
// 
//  v3  v2
//   
//  v0  v1 
//

void MMCell::drawGL()
{
	if(selected) glColor3f(1,0,0);
	else glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glNormal(n);
	glTexCoord2f(0.05f,0.05f);
	glVertex(V[0]);
	glTexCoord2f(0.95f,0.05f);
	glVertex(V[1]);
	glTexCoord2f(0.95f,0.95f);
	glVertex(V[2]);
	glTexCoord2f(0.05f,0.95f);
	glVertex(V[3]);
	
	glNormal3f(0,-1,0);
	glVertex  (V[0]);
	glVertex3f(V[0][0], V[0][1], 0);
	glVertex3f(V[1][0], V[1][1], 0);
	glVertex  (V[1]);
	
  glNormal3f(1,0,0);
	glVertex  (V[1]);
	glVertex3f(V[1][0], V[1][1], 0);
	glVertex3f(V[2][0], V[2][1], 0);
	glVertex  (V[2]);

	glNormal3f(0,1,0);
	glVertex  (V[2]);
	glVertex3f(V[2][0], V[2][1], 0);
	glVertex3f(V[3][0], V[3][1], 0);
	glVertex  (V[3]);

	glNormal3f(-1,0,0);
	glVertex  (V[3]);
	glVertex3f(V[3][0], V[3][1], 0);
	glVertex3f(V[0][0], V[0][1], 0);
	glVertex  (V[0]);

	glEnd();

}

void MMCell::ComputeNormal()
{
	Point3f e01=V[1]-V[0];
	Point3f e02=V[2]-V[0];
	n=e01^e02;
	n.Normalize();
}

void MMGrid::initGL()
{
		glEnable(GL_TEXTURE_2D);
			glGenTextures(1, &ti);
			QImage tt; 	tt.load(texFileName);
			QImage tx = QGLWidget::convertToGLFormat ( tt);
			glBindTexture(GL_TEXTURE_2D, ti);
			glTexImage2D( GL_TEXTURE_2D, 0, 3, tx.width(), tx.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, tx.bits() );
			gluBuild2DMipmaps(GL_TEXTURE_2D, 3, tx.width(), tx.height(), GL_RGBA, GL_UNSIGNED_BYTE, tx.bits() );
}

void MMGrid::clearSelection()
{
	vector<MMCell>::iterator ii;
	for (ii=g.begin();ii!=g.end();++ii) (*ii).selected=false;
 }

void MMGrid::selectRect(Point2i s, Point2i e)
{
	int i,j;
	if(s[0]>e[0]) swap(s[0],e[0]);
	if(s[1]>e[1]) swap(s[1],e[1]);
	for(i=s[0];i<=e[0];++i)
		for(j=s[1];j<=e[1];++j)
			G(i,j).selected=true;	
}

void MMGrid::raiseSelected(float f)
{
	vector<MMCell>::iterator ii;
	for (ii=g.begin();ii!=g.end();++ii) 
		if((*ii).selected)
			for(int k=0;k<4;++k)
				(*ii).V[k].z()+=f;
}

void MMGrid::smoothSelected()
{
	int bp[4][2]={{0,0},{1,0},{1,1},{0,1}}; // base position of each vertex of the cell
	int dp[4][2]={{-1,-1},{0,-1},{0,0},{-1,0}}; // delta of the four adj of a cell vertex
	int i,j,k,q;
	int cnt;
	float tH;
	for(i=0;i<sx;++i)
		for(j=0;j<sy;++j)
			if(G(i,j).selected)
				for(k=0;k<4;++k) //for each cell vertex;
				{
					tH=0;
					cnt=0;
					for(q=0;q<4;++q) // foreach adj of a cell vertexe
						if(G(i+bp[k][0]+dp[q][0],j+bp[k][1]+dp[q][1]).selected)
						{
							tH+=G(i+bp[k][0]+dp[q][0],j+bp[k][1]+dp[q][1]).avgH();
							++cnt;
						}
					G(i,j).tmpH[k]=tH/cnt;
				}

	vector<MMCell>::iterator ii;
	for (ii=g.begin();ii!=g.end();++ii) 
		if((*ii).selected)
				for(int k=0;k<4;++k) 
					(*ii).V[k][2]=(*ii).tmpH[k];
}

void MMGrid::flattenSelected()
{
	// first step compute average height
	int i,j;
	int cnt=0;
	float avgH=0;
	for(i=0;i<sx;++i)
		for(j=0;j<sy;++j)
			if(G(i,j).selected)	{
				avgH+=G(i,j).avgH();
				++cnt;
			}
	avgH/=cnt;
	// second step apply the compute average height
	vector<MMCell>::iterator ii;
	for (ii=g.begin();ii!=g.end();++ii) 
		if((*ii).selected)			
			for(int k=0;k<4;++k) 
					(*ii).V[k][2]=avgH;
}


void MMFace::ComputeE()
	{	
			// Primo calcolo degli edges
		edge[0] = (*(v[1])); edge[0] -= (*(v[0]));
		edge[1] = (*(v[2])); edge[1] -= (*(v[1]));
		edge[2] = (*(v[0])); edge[2] -= (*(v[2]));
			// Calcolo di plane
		plane.n = edge[0]^edge[1];
		plane.d = plane.n * (*(v[0]));
		plane.Normalize();
			// Calcolo migliore proiezione
		float nx = Abs(plane.n[0]);
		float ny = Abs(plane.n[1]);
		float nz = Abs(plane.n[2]);
		float d;
		if(nx>ny && nx>nz) { flags |= NORMX; d = 1/plane.n[0]; }
		else if(ny>nz)     { flags |= NORMY; d = 1/plane.n[1]; }
		else               { flags |= NORMZ; d = 1/plane.n[2]; }

			// Scalatura spigoli
		edge[0] *= d;
		edge[1] *= d;
		edge[2] *= d;
	}

bool MMFace::Dist( const Point3f & q, float & dist, Point3f & p )
	{
		const float EPSILON = float(- 0.000001);
		float b;
			// Calcolo distanza punto piano
		float d = Distance( plane, q );
		if( d>dist || d<-dist )			// Risultato peggiore: niente di fatto
			return false;

			// Calcolo del punto sul piano
		// NOTA: aggiunto un '-d' in fondo Paolo C.
		Point3f t = plane.n;
		t[0] *= -d;
		t[1] *= -d;
		t[2] *= -d;
		p = q; p += t;
		
	#define PP(i)	(*(v[i]))
	#define E(i)    (edge[i])
    
		switch( flags & (NORMX|NORMY|NORMZ) )
		{
		case NORMX:
			b = E(1)[1]*(p[2] - PP(1)[2]) - E(1)[2]*(p[1] - PP(1)[1]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[1])),(*(v[2])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(2)[1]*(p[2] - PP(2)[2]) - E(2)[2]*(p[1] - PP(2)[1]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[2])),(*(v[0])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(0)[1]*(p[2] - PP(0)[2]) - E(0)[2]*(p[1] - PP(0)[1]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[0])),(*(v[1])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			break;

		  case NORMY:
			b = E(1)[2]*(p[0] - PP(1)[0]) - E(1)[0]*(p[2] - PP(1)[2]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[1])),(*(v[2])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(2)[2]*(p[0] - PP(2)[0]) - E(2)[0]*(p[2] - PP(2)[2]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[2])),(*(v[0])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(0)[2]*(p[0] - PP(0)[0]) - E(0)[0]*(p[2] - PP(0)[2]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[0])),(*(v[1])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			break;

		  case NORMZ:
			b = E(1)[0]*(p[1] - PP(1)[1]) - E(1)[1]*(p[0] - PP(1)[0]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[1])),(*(v[2])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(2)[0]*(p[1] - PP(2)[1]) - E(2)[1]*(p[0] - PP(2)[0]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[2])),(*(v[0])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			b = E(0)[0]*(p[1] - PP(0)[1]) - E(0)[1]*(p[0] - PP(0)[0]);
			if(b<-EPSILON)
			{
				b = PSDist(q,(*(v[0])),(*(v[1])),p);
				if(dist>b) { dist = b; return true; }
				else return false;
			}
			break;
		}

	#undef E
	#undef PP

		dist = fabs(d);
		//dist = Distance(p,q);
		return true;
	}




// Generate Face and cell
// si genera un insieme di vertici minimo
void MMCell::generateFace()
{   int k;
	for(k=0;k<4;++k) B[k]=Point3f(V[k][0],V[k][1],0);

	face.push_back(MMFace(&(V[0]),&(V[1]),&(V[2])));
	face.push_back(MMFace(&(V[0]),&(V[2]),&(V[3])));

	face.push_back(MMFace(&(V[0]),&(B[0]),&(V[1])));
	face.push_back(MMFace(&(V[1]),&(B[0]),&(B[1])));
	face.push_back(MMFace(&(V[1]),&(B[1]),&(V[2])));
	face.push_back(MMFace(&(V[2]),&(B[1]),&(B[2])));
	face.push_back(MMFace(&(V[2]),&(B[2]),&(V[3])));
	face.push_back(MMFace(&(V[3]),&(B[2]),&(B[3])));
	face.push_back(MMFace(&(V[3]),&(B[3]),&(V[0])));
	face.push_back(MMFace(&(V[0]),&(B[3]),&(B[0])));
}
	
		
MMFace::MMFace(Point3f * v0, Point3f * v1, Point3f * v2)
{
	v[0]=v0;
	v[1]=v1;
	v[2]=v2;
	ComputeE();
}
