#include <GL/glew.h>
#include "CI/CISG.h"
#include "CI/CICubeMap.h"
#include "CI/CIMsg.h"
#include "MMGrid.h"
#include "MMSG.h"

using namespace vcg ;

void CICubeMapNode::XMLWrite(QTextStream &xstrm)
{
	xstrm << "<CICubeMapNode>\n";
	xstrm << textureBaseName ;
	xstrm << "</CICubeMapNode>\n";
}

bool CICubeMapNode::XMLRead(QDomElement n, CISG *) 
{
	if(n.tagName() !="CICubeMapNode" || !n.firstChild().isText() ) 
		return false;
	qDebug("< %s > %s ...\n",n.tagName().ascii(),n.text().ascii());
	QTextIStream(&n.text()) >> textureBaseName;
  return true;
}

void CICubeMapNode::drawGL(const float )
{
	if(dirty) initializeGL();
	CIMsg::checkGLError("drawGL");
	// estrae la componente rotazione dalla modelview corrente.
	// 
	Matrix44f mm;
	glGetModelViewMatrix(mm);
	mm.a[0][3]=0;
	mm.a[1][3]=0;
	mm.a[2][3]=0;
	mm.a[3][3]=1;
	float dt=1.0f/pow(mm.Det(),1.0/3.0);
	Matrix44f ss; ss.Scale(dt,dt,dt);
	rr=mm*ss;
	dt=rr.Det();
	DrawEnvCube(rr);
	glDisable(GL_TEXTURE_2D);
}


/********** CISGGround *************/	
void MMSGSphere::drawGL(const float )
{
	if(!dl) 
	{
	 dl=glGenLists(1);
	 glNewList(dl,GL_COMPILE);
	 GLUquadricObj *qobj=gluNewQuadric();
     gluSphere(qobj,1,12,6);
	 glEndList();
    }
	glCallList(dl);
}
void MMSGSphere::XMLWrite(QTextStream &xstrm)
{
	xstrm << "<CISGSphere>\n";
	xstrm << "</CISGSphere>\n";
}
bool MMSGSphere::XMLRead(QDomElement n, CISG *)
{
	if(n.tagName() !="CISGSphere" || !n.isText()) return false;
	return true;
}



void CIGridNode::XMLWrite(QTextStream &xstrm)
{
	xstrm << "<CIGridNode>\n";
	xstrm << "</CIGridNode>\n";
}
bool CIGridNode::XMLRead(QDomElement en, CISG *) 
{
	return true;
}


CISGNode *MMSG::Allocate(const QString &classname)
{
	CISGNode *ret=CISG::Allocate(classname);		
	if(classname=="CIGrid") ret=new CIGridNode();
	if(classname=="CICubeMap") ret=new CICubeMapNode();
	
	return ret;
}	


